from typing import Literal, TypedDict

from omu.interface.keyable import Keyable
from omu.model import Model

type ArchiveStatus = Literal["pending", "processing", "completed", "failed"]


class ArchiveData(TypedDict):
    id: str
    url: str
    title: str | None
    description: str | None
    thumbnail: str | None
    published_at: str | None
    duration: int | None
    status: ArchiveStatus


class Archive(Model[ArchiveData], Keyable):
    def __init__(self, data: ArchiveData):
        self.id = data["id"]
        self.url = data["url"]
        self.title = data["title"]
        self.description = data["description"]
        self.thumbnail = data["thumbnail"]
        self.published_at = data["published_at"]
        self.duration = data["duration"]
        self.status: ArchiveStatus = data["status"]

    @classmethod
    def from_json(cls, json: ArchiveData) -> "Archive":
        return cls(json)

    def to_json(self) -> ArchiveData:
        return {
            "id": self.id,
            "url": self.url,
            "title": self.title,
            "description": self.description,
            "thumbnail": self.thumbnail,
            "published_at": self.published_at,
            "duration": self.duration,
            "status": self.status,
        }

    def key(self) -> str:
        return self.id


class YtDlpInfo(TypedDict):
    version: str
    git_head: str
    variant: str
    update_hint: str
    channel: str
    origin: str


class ArchiveLimit(TypedDict):
    size_mb: int
    count: int
    duration_days: int


class ArchiveConfig(TypedDict):
    active: bool
    yt_dlp_info: YtDlpInfo
    yt_dlp_options: dict[str, str | bool | int | list[int] | None]
    output_dir: str
    archive_limit: ArchiveLimit



from omu.extension.endpoint.endpoint import EndpointType
from omu.extension.registry import RegistryType
from omu.extension.table import TableType

from .archive import Archive, ArchiveConfig
from .const import DEFAULT_CONFIG, IDENTIFIER

ARCHIVE_TABLE_TYPE = TableType.create_model(
    IDENTIFIER,
    "archive",
    Archive,
)
CONFIG_REGISTRY_TYPE = RegistryType[ArchiveConfig].create_json(
    IDENTIFIER,
    "config",
    DEFAULT_CONFIG,
)
OPEN_OUTPUT_DIR_ENDPOINT_TYPE = EndpointType[None, None].create_json(
    IDENTIFIER,
    "open_output_dir",
)



import asyncio
import os
from pathlib import Path

import yt_dlp.version
from omu import Omu
from omu_chat import Chat, Room, events

from .archive import Archive, ArchiveConfig
from .const import APP
from .types import (
    ARCHIVE_TABLE_TYPE,
    CONFIG_REGISTRY_TYPE,
    OPEN_OUTPUT_DIR_ENDPOINT_TYPE,
)

omu = Omu(APP)
chat = Chat(omu)

archive_table = omu.tables.get(ARCHIVE_TABLE_TYPE)
config_registry = omu.registries.get(CONFIG_REGISTRY_TYPE)


@config_registry.listen
async def on_config_update(config: ArchiveConfig):
    path = Path(config["output_dir"])
    path.mkdir(parents=True, exist_ok=True)
    if config["active"]:
        rooms = await chat.rooms.fetch_items(limit=10, backward=True)
        for room in rooms.values():
            await process_room(room)


@omu.endpoints.bind(endpoint_type=OPEN_OUTPUT_DIR_ENDPOINT_TYPE)
async def handle_open_output_dir(req: None):
    os.startfile(config_registry.value["output_dir"])


archive_threads: dict[str, Archive] = {}


def archive_thread(archive: Archive):
    options = {
        **config_registry.value["yt_dlp_options"],
        "paths": {
            "home": config_registry.value["output_dir"],
        },
    }
    with yt_dlp.YoutubeDL(options) as ydl:
        info = ydl.extract_info(archive.url, download=False)
        if info:
            archive.description = info.get("description", archive.description)
            archive.duration = info.get("duration", archive.duration)
            archive.published_at = info.get("published_at", archive.published_at)
            archive.thumbnail = info.get("thumbnail", archive.thumbnail)
            archive.title = info.get("title", archive.title)
        ydl.download([archive.url])


async def start_archive(archive: Archive):
    if archive.id in archive_threads:
        return
    archive_threads[archive.id] = archive

    def thread():
        try:
            archive_thread(archive)
            archive.status = "completed"
        except Exception:
            archive.status = "failed"
        finally:
            del archive_threads[archive.id]
            asyncio.run_coroutine_threadsafe(archive_table.update(archive), omu.loop)

    asyncio.get_running_loop().run_in_executor(None, thread)


@chat.on(events.room.add)
async def on_room_add(room: Room):
    if not config_registry.value["active"]:
        return
    return await process_room(room)


async def process_room(room: Room):
    metadata = room.metadata or {}
    url = metadata.get("url")
    if url is None:
        return
    archive = Archive(
        {
            "id": room.id.key(),
            "url": url,
            "status": "pending",
            "description": metadata.get("description"),
            "duration": metadata.get("duration"),
            "published_at": metadata.get("published_at"),
            "thumbnail": metadata.get("thumbnail"),
            "title": metadata.get("title"),
        }
    )
    await archive_table.add(archive)
    await start_archive(archive)


@chat.on(events.room.update)
async def on_room_update(room: Room):
    archive = await archive_table.get(room.id.key())
    if archive is None:
        return
    metadata = room.metadata or {}
    archive.description = metadata.get("description", archive.description)
    archive.published_at = metadata.get("created_at", archive.published_at)
    archive.thumbnail = metadata.get("thumbnail", archive.thumbnail)
    archive.title = metadata.get("title", archive.title)
    await archive_table.update(archive)


async def refresh_ytdlp_info():
    @config_registry.update
    async def update_config(config: ArchiveConfig) -> ArchiveConfig:
        return {
            **config,
            "yt_dlp_info": {
                "version": yt_dlp.version.__version__,
                "git_head": yt_dlp.version.RELEASE_GIT_HEAD,
                "variant": yt_dlp.version.VARIANT,
                "update_hint": yt_dlp.version.UPDATE_HINT,
                "channel": yt_dlp.version.CHANNEL,
                "origin": yt_dlp.version.ORIGIN,
            },
        }

    await update_config


async def process_pending_archives():
    archive_records = await archive_table.fetch_items(limit=10, backward=True)
    for archive in archive_records.values():
        if archive.status != "pending":
            continue
        await start_archive(archive)


@omu.on_ready
async def on_ready():
    await refresh_ytdlp_info()
    await process_pending_archives()



from pathlib import Path

import yt_dlp.version
from omu.app import App
from omu.identifier import Identifier

from .archive import ArchiveConfig
from .version import VERSION

IDENTIFIER = Identifier.from_key("com.omuapps:archive/plugin")
APP = App(
    id=IDENTIFIER,
    version=VERSION,
)

# yt-dlp <url> --live-from-start --wait-for-video 60 --write-thumbnail --write-info-json --write-description --write-annotations --write-sub
# --live-from-start   - Start the stream from the beginning
# --wait-for-video    - Wait for video
# --write-thumbnail   - Write thumbnail image
# --write-info-json   - Write video metadata
# --write-description - Write video description
# --write-annotations - Write video annotations
# --write-sub         - Write video subtitles
# -P "C:/MyVideos"    - Save the video to the specified directory
YTDLP_OPTIONS = {
    "live_from_start": True,
    "wait_for_video": [60, 60],
    "writethumbnail": True,
    "writeinfojson": True,
    "writedescription": True,
    "writeannotations": True,
    # "writesubtitles": True,
}

DEFAULT_CONFIG = ArchiveConfig(
    active=False,
    yt_dlp_info={
        "version": yt_dlp.version.__version__,
        "git_head": yt_dlp.version.RELEASE_GIT_HEAD,
        "variant": yt_dlp.version.VARIANT,
        "update_hint": yt_dlp.version.UPDATE_HINT,
        "channel": yt_dlp.version.CHANNEL,
        "origin": yt_dlp.version.ORIGIN,
    },
    yt_dlp_options=YTDLP_OPTIONS,
    output_dir=str(Path.home() / "Videos" / "omuapps.com" / "archive"),
    archive_limit={
        "size_mb": 1024 * 16,
        "count": 100,
        "duration_days": 30,
    },
)
